from __future__ import annotations

import datetime
import typing
import uuid
from typing import Any

from pydantic import BaseModel
from pydantic.fields import ModelPrivateAttr

from core.util import date_util


class Message(BaseModel):  # type: ignore[explicit-any]
    command: str
    content: dict[str, Any]  # type: ignore[explicit-any]
    requestId: str | None
    postCount: int | None
    postDate: datetime.datetime | None

    def prepare_for_send(self) -> None:
        self.requestId = self.requestId or str(uuid.uuid4()).replace('-', '')
        self.postCount = (self.postCount + 1) if self.postCount else 1
        self.postDate = date_util.datetime_from_now()


class MessageContent(BaseModel):
    _COMMAND = 'UNKNOWN_COMMAND'

    @classmethod
    def get_command(cls) -> str:
        return typing.cast(str, typing.cast(ModelPrivateAttr, cls._COMMAND).get_default())

    def to_message(self) -> Message:
        return Message(
            command=self.get_command(),
            content=self.model_dump(),
            requestId=None,
            postCount=None,
            postDate=None,
        )



from core.notifications.notification_client import NotificationClient
from core.requester import KibaResponse
from core.requester import Requester


class DiscordClient(NotificationClient):
    def __init__(self, webhookUrl: str, requester: Requester) -> None:
        self.webhookUrl = webhookUrl
        self.requester = requester

    async def post(self, messageText: str) -> KibaResponse:
        data = {'content': messageText}
        response = await self.requester.post(self.webhookUrl, dataDict=data, headers={'content-type': 'application/json'})
        return response



import datetime
import hashlib
import hmac
import json
from collections.abc import Mapping
from collections.abc import MutableMapping
from collections.abc import Sequence
from typing import Union
from urllib import parse as urlparse

from core.exceptions import InternalServerErrorException
from core.requester import FileContent
from core.requester import HttpxFileTypes
from core.requester import KibaResponse
from core.requester import Requester
from core.util import date_util
from core.util.typing_util import Json


# NOTE(krishan711): mostly adapted from https://docs.aws.amazon.com/general/latest/gr/sigv4-signed-request-examples.html
class AwsRequester(Requester):
    _SIGNING_ALGORITHM = 'AWS4-HMAC-SHA256'

    def __init__(self, accessKeyId: str, accessKeySecret: str, headers: dict[str, str] | None = None) -> None:
        super().__init__(headers=headers)
        self.accessKeyId = accessKeyId
        self.accessKeySecret = accessKeySecret

    @staticmethod
    def _sign(key: bytes, message: str) -> bytes:
        return hmac.new(key, message.encode(), hashlib.sha256).digest()

    @staticmethod
    def _sign_hex(key: bytes, message: str) -> str:
        return hmac.new(key, message.encode(), hashlib.sha256).hexdigest()

    def _sign_string(self, stringToSign: str, requestDate: datetime.datetime, service: str, region: str) -> str:
        key1 = f'AWS4{self.accessKeySecret}'.encode()
        key2 = self._sign(key=key1, message=requestDate.strftime('%Y%m%d'))
        key3 = self._sign(key=key2, message=region)
        key4 = self._sign(key=key3, message=service)
        key5 = self._sign(key=key4, message='aws4_request')
        return self._sign_hex(key=key5, message=stringToSign)

    async def make_request(
        self,
        method: str,
        url: str,
        dataDict: Json | None = None,
        data: bytes | None = None,
        formDataDict: Mapping[str, Union[str, FileContent]] | None = None,
        formFiles: Sequence[tuple[str, HttpxFileTypes]] | None = None,
        timeout: int | None = 10,
        headers: MutableMapping[str, str] | None = None,
        outputFilePath: str | None = None,
    ) -> KibaResponse:
        canonicalQueryString = ''
        if data is None and dataDict is not None:
            if method == 'GET':
                raise InternalServerErrorException('GET requests with parameters are not supported on AwsRequester yet.')
            if method == 'POST':
                data = json.dumps(dataDict).encode()
        if not data:
            raise InternalServerErrorException('requests without date are not supported on AwsRequester yet.')
        if formDataDict:
            raise InternalServerErrorException('formDataDict is not supported on AwsRequester yet.')
        if formFiles:
            raise InternalServerErrorException('formFiles is not supported on AwsRequester yet.')
        parsedUrl = urlparse.urlparse(url=url)
        host = parsedUrl.netloc
        path = parsedUrl.path or '/'
        hostParts = host.split('.')
        service = hostParts[-4]
        region = hostParts[-3]
        requestDate = date_util.datetime_from_now()
        amazonFormattedDate = requestDate.strftime('%Y%m%dT%H%M%SZ')
        headers = headers or {}
        headers['x-amz-date'] = amazonFormattedDate
        headers['host'] = host
        # NOTE(krishan711): headers always have to be sorted before signing
        headerKeysToSign = sorted(['host', 'x-amz-date'])
        signedHeadersString = ';'.join(headerKeysToSign)
        canonicalHeaders = ''.join(f'{headerKey}:{headers[headerKey]}\n' for headerKey in headerKeysToSign)
        payloadHash = hashlib.sha256(data).hexdigest()
        canonicalRequest = '\n'.join([method, path, canonicalQueryString, canonicalHeaders, signedHeadersString, payloadHash])  # noqa: FLY002
        requestHash = hashlib.sha256(canonicalRequest.encode()).hexdigest()
        credentialScope = '/'.join([requestDate.strftime('%Y%m%d'), region, service, 'aws4_request'])
        stringToSign = '\n'.join([self._SIGNING_ALGORITHM, amazonFormattedDate, credentialScope, requestHash])  # noqa: FLY002
        signature = self._sign_string(stringToSign=stringToSign, requestDate=requestDate, service=service, region=region)
        headers['Authorization'] = f'{self._SIGNING_ALGORITHM} Credential={self.accessKeyId}/{credentialScope}, SignedHeaders={signedHeadersString}, Signature={signature}'
        return await super().make_request(method=method, url=url, dataDict=None, data=data, formDataDict=None, formFiles=None, timeout=timeout, headers=headers, outputFilePath=outputFilePath)



import dataclasses
import datetime
from collections.abc import Sequence
from enum import Enum

import sqlalchemy
from sqlalchemy import Table
from sqlalchemy.sql import Select

from core.store.database import Database
from core.store.database import ResultType


class Direction(Enum):
    ASCENDING = 'ascending'
    DESCENDING = 'descending'


@dataclasses.dataclass
class Order:
    fieldName: str
    direction: Direction = Direction.DESCENDING


@dataclasses.dataclass
class RandomOrder(Order):
    fieldName: str = '__KIBA_RANDOM'


@dataclasses.dataclass
class FieldFilter:
    fieldName: str
    isNull: bool | None = None
    isNotNull: bool | None = None


@dataclasses.dataclass
class StringFieldFilter(FieldFilter):
    eq: str | None = None
    ne: str | None = None
    containedIn: Sequence[str] | None = None
    notContainedIn: Sequence[str] | None = None


@dataclasses.dataclass
class DateFieldFilter(FieldFilter):
    eq: datetime.datetime | None = None
    ne: datetime.datetime | None = None
    lte: datetime.datetime | None = None
    lt: datetime.datetime | None = None
    gte: datetime.datetime | None = None
    gt: datetime.datetime | None = None
    containedIn: Sequence[datetime.datetime] | None = None
    notContainedIn: Sequence[datetime.datetime] | None = None


@dataclasses.dataclass
class IntegerFieldFilter(FieldFilter):
    eq: int | None = None
    ne: int | None = None
    lte: int | None = None
    lt: int | None = None
    gte: int | None = None
    gt: int | None = None
    containedIn: Sequence[int] | None = None
    notContainedIn: Sequence[int] | None = None


@dataclasses.dataclass
class FloatFieldFilter(FieldFilter):
    eq: float | None = None
    ne: float | None = None
    lte: float | None = None
    lt: float | None = None
    gte: float | None = None
    gt: float | None = None
    containedIn: Sequence[float] | None = None
    notContainedIn: Sequence[float] | None = None


@dataclasses.dataclass
class BooleanFieldFilter(FieldFilter):
    eq: bool | None = None
    ne: bool | None = None


class Retriever:
    def __init__(self, database: Database) -> None:
        self.database = database

    def _apply_order(self, query: Select[ResultType], table: Table, order: Order) -> Select[ResultType]:
        if isinstance(order, RandomOrder):
            query = query.order_by(sqlalchemy.sql.functions.random())
        else:
            field = table.c[order.fieldName]
            query = query.order_by(field.asc() if order.direction == Direction.ASCENDING else field.desc())
        return query

    def _apply_orders(self, query: Select[ResultType], table: Table, orders: Sequence[Order]) -> Select[ResultType]:
        for order in orders:
            query = self._apply_order(query=query, table=table, order=order)
        return query

    def _apply_string_field_filter(self, query: Select[ResultType], table: Table, fieldFilter: StringFieldFilter) -> Select[ResultType]:
        field = table.c[fieldFilter.fieldName]
        if fieldFilter.eq is not None:
            query = query.where(field == fieldFilter.eq)
        if fieldFilter.ne is not None:
            query = query.where(field != fieldFilter.ne)
        if fieldFilter.containedIn is not None:
            query = query.where(field.in_(fieldFilter.containedIn))
        if fieldFilter.notContainedIn is not None:
            query = query.where(field.not_in(fieldFilter.notContainedIn))
        return query

    def _apply_date_field_filter(self, query: Select[ResultType], table: Table, fieldFilter: DateFieldFilter) -> Select[ResultType]:
        field = table.c[fieldFilter.fieldName]
        if fieldFilter.eq is not None:
            query = query.where(field == fieldFilter.eq)
        if fieldFilter.ne is not None:
            query = query.where(field != fieldFilter.ne)
        if fieldFilter.lte is not None:
            query = query.where(field <= fieldFilter.lte)
        if fieldFilter.lt is not None:
            query = query.where(field < fieldFilter.lt)
        if fieldFilter.gte is not None:
            query = query.where(field >= fieldFilter.gte)
        if fieldFilter.gt is not None:
            query = query.where(field > fieldFilter.gt)
        if fieldFilter.containedIn is not None:
            query = query.where(field.in_(fieldFilter.containedIn))
        if fieldFilter.notContainedIn is not None:
            query = query.where(field.not_in(fieldFilter.notContainedIn))
        return query

    def _apply_integer_field_filter(self, query: Select[ResultType], table: Table, fieldFilter: IntegerFieldFilter) -> Select[ResultType]:
        field = table.c[fieldFilter.fieldName]
        if fieldFilter.eq is not None:
            query = query.where(field == fieldFilter.eq)
        if fieldFilter.ne is not None:
            query = query.where(field != fieldFilter.ne)
        if fieldFilter.lte is not None:
            query = query.where(field <= fieldFilter.lte)
        if fieldFilter.lt is not None:
            query = query.where(field < fieldFilter.lt)
        if fieldFilter.gte is not None:
            query = query.where(field >= fieldFilter.gte)
        if fieldFilter.gt is not None:
            query = query.where(field > fieldFilter.gt)
        if fieldFilter.containedIn is not None:
            query = query.where(field.in_(fieldFilter.containedIn))
        if fieldFilter.notContainedIn is not None:
            query = query.where(field.not_in(fieldFilter.notContainedIn))
        return query

    def _apply_float_field_filter(self, query: Select[ResultType], table: Table, fieldFilter: FloatFieldFilter) -> Select[ResultType]:
        field = table.c[fieldFilter.fieldName]
        if fieldFilter.eq is not None:
            query = query.where(field == fieldFilter.eq)
        if fieldFilter.ne is not None:
            query = query.where(field != fieldFilter.ne)
        if fieldFilter.lte is not None:
            query = query.where(field <= fieldFilter.lte)
        if fieldFilter.lt is not None:
            query = query.where(field < fieldFilter.lt)
        if fieldFilter.gte is not None:
            query = query.where(field >= fieldFilter.gte)
        if fieldFilter.gt is not None:
            query = query.where(field > fieldFilter.gt)
        if fieldFilter.containedIn is not None:
            query = query.where(field.in_(fieldFilter.containedIn))
        if fieldFilter.notContainedIn is not None:
            query = query.where(field.not_in(fieldFilter.notContainedIn))
        return query

    def _apply_boolean_field_filter(self, query: Select[ResultType], table: Table, fieldFilter: BooleanFieldFilter) -> Select[ResultType]:
        field = table.c[fieldFilter.fieldName]
        if fieldFilter.eq is not None:
            query = query.where(field == fieldFilter.eq)
        if fieldFilter.ne is not None:
            query = query.where(field != fieldFilter.ne)
        return query

    def _apply_field_filter(self, query: Select[ResultType], table: Table, fieldFilter: FieldFilter) -> Select[ResultType]:
        field = table.c[fieldFilter.fieldName]
        if fieldFilter.isNull:
            query = query.where(field.is_(None))
        if fieldFilter.isNotNull:
            query = query.where(field.is_not(None))
        if isinstance(fieldFilter, StringFieldFilter):
            query = self._apply_string_field_filter(query=query, table=table, fieldFilter=fieldFilter)
        if isinstance(fieldFilter, DateFieldFilter):
            query = self._apply_date_field_filter(query=query, table=table, fieldFilter=fieldFilter)
        if isinstance(fieldFilter, IntegerFieldFilter):
            query = self._apply_integer_field_filter(query=query, table=table, fieldFilter=fieldFilter)
        if isinstance(fieldFilter, FloatFieldFilter):
            query = self._apply_float_field_filter(query=query, table=table, fieldFilter=fieldFilter)
        if isinstance(fieldFilter, BooleanFieldFilter):
            query = self._apply_boolean_field_filter(query=query, table=table, fieldFilter=fieldFilter)
        return query

    def _apply_field_filters(self, query: Select[ResultType], table: Table, fieldFilters: Sequence[FieldFilter]) -> Select[ResultType]:
        for fieldFilter in fieldFilters:
            query = self._apply_field_filter(query=query, table=table, fieldFilter=fieldFilter)
        return query



import os
import tempfile
import shutil

import pytest

from core.caching.file_cache import FileCache
from core.util import date_util


class TestFileCache:

    @pytest.fixture
    def cache_dir(self):
        # Create a temporary directory for testing
        temp_dir = tempfile.mkdtemp()
        yield temp_dir
        # Clean up after tests
        shutil.rmtree(temp_dir)

    @pytest.fixture
    def cache(self, cache_dir) -> FileCache:
        return FileCache(cacheDirectory=cache_dir)

    async def test_set_and_get(self, cache: FileCache):
        key = "test_key"
        value = "test_value"
        success = await cache.set(key=key, value=value, expirySeconds=60)
        assert success is True
        result = await cache.get(key=key)
        assert result == value

    async def test_get_nonexistent_key(self, cache: FileCache):
        result = await cache.get(key="nonexistent")
        assert result is None

    async def test_delete_existing_key(self, cache: FileCache):
        key = "test_key"
        value = "test_value"
        await cache.set(key=key, value=value, expirySeconds=60)
        success = await cache.delete(key=key)
        assert success is True
        result = await cache.get(key=key)
        assert result is None

    async def test_delete_nonexistent_key(self, cache: FileCache):
        success = await cache.delete(key="nonexistent")
        assert success is False

    async def test_expiry(self, cache: FileCache, monkeypatch):
        key = "test_key"
        value = "test_value"
        await cache.set(key=key, value=value, expirySeconds=60)
        future_time = date_util.datetime_from_now(seconds=61)
        monkeypatch.setattr(date_util, "datetime_from_now", lambda: future_time)
        result = await cache.get(key=key)
        assert result is None

    async def test_overwrite_existing_key(self, cache: FileCache):
        key = "test_key"
        value1 = "test_value_1"
        value2 = "test_value_2"
        await cache.set(key=key, value=value1, expirySeconds=60)
        await cache.set(key=key, value=value2, expirySeconds=60)
        result = await cache.get(key=key)
        assert result == value2

    def test_can_store_complex_objects(self, cache: FileCache):
        assert cache.can_store_complex_objects() is False

    def test_is_private_flag(self, cache_dir):
        private_cache = FileCache(cacheDirectory=cache_dir, isPrivate=True)
        assert private_cache.isPrivate is True
        public_cache = FileCache(cacheDirectory=cache_dir, isPrivate=False)
        assert public_cache.isPrivate is False

    async def test_file_structure(self, cache: FileCache, cache_dir):
        key = "test_key"
        value = "test_value"
        await cache.set(key=key, value=value, expirySeconds=60)
        cache_key_dir = os.path.join(cache_dir, key)
        content_file = os.path.join(cache_key_dir, 'content.txt')
        expiry_file = os.path.join(cache_key_dir, 'expiryDate.txt')
        assert os.path.exists(cache_key_dir)
        assert os.path.isfile(content_file)
        assert os.path.isfile(expiry_file)
        with open(content_file, 'r') as f:
            assert f.read() == value

    async def test_missing_expiry_file(self, cache: FileCache, cache_dir):
        key = "test_key"
        value = "test_value"
        await cache.set(key=key, value=value, expirySeconds=60)
        expiry_file = os.path.join(cache_dir, key, 'expiryDate.txt')
        os.remove(expiry_file)
        result = await cache.get(key=key)
        assert result is None
        content_file = os.path.join(cache_dir, key, 'content.txt')
        assert not os.path.exists(content_file)



from collections.abc import Mapping
from typing import Any
from typing import Union

JsonBaseType = str | int | float | bool | None

JSON5 = JsonBaseType
JSON4 = Union[JsonBaseType, list[JSON5], Mapping[str, JSON5]]
JSON3 = Union[JsonBaseType, list[JSON4], Mapping[str, JSON4]]
JSON2 = Union[JsonBaseType, list[JSON3], Mapping[str, JSON3]]
JSON1 = Union[JsonBaseType, list[JSON2], Mapping[str, JSON2]]
JSON = Union[JsonBaseType, list[JSON1], Mapping[str, JSON1]]
UnsafeJSON5 = Union[JsonBaseType, list[Any], Mapping[str, Any]]  # type: ignore[explicit-any]
UnsafeJSON4 = Union[JsonBaseType, list[UnsafeJSON5], Mapping[str, UnsafeJSON5]]
UnsafeJSON3 = Union[JsonBaseType, list[UnsafeJSON4], Mapping[str, UnsafeJSON4]]
UnsafeJSON2 = Union[JsonBaseType, list[UnsafeJSON3], Mapping[str, UnsafeJSON3]]
UnsafeJSON1 = Union[JsonBaseType, list[UnsafeJSON2], Mapping[str, UnsafeJSON2]]
UnsafeJSON = Union[JsonBaseType, list[UnsafeJSON1], Mapping[str, UnsafeJSON1]]

# NOTE(krishan711): copied from https://github.com/python/typing/issues/182
# NOTE(krishan711): this doesn't work great with pydantic yet due to: https://github.com/pydantic/pydantic/issues/7111
# NOTE(krishan711): Json: TypeAlias = dict[str, "Json"] | Mapping[str, "Json"] | list["Json"] | Sequence["Json"] | JsonBaseType
Json = JSON
JsonDict = dict[str, Json]
JsonList = list[Json]



import os
import pathlib
import shutil
import time

import aiofiles
import aiofiles.os

KILOBYTE = 1024
MEGABYTE = KILOBYTE * 1024

CACHE_CONTROL_TEMPORARY_FILE = 'public,max-age=1'
CACHE_CONTROL_FINAL_FILE = 'public,max-age=31536000'


async def remove_file(filePath: str) -> None:
    await aiofiles.os.remove(filePath)


async def remove_directory(directory: str) -> None:
    # TODO(krish): fix this to be async, command below doesn't work if the directory is not empty
    shutil.rmtree(directory)


def remove_file_sync(filePath: str) -> None:
    os.remove(filePath)


def remove_directory_sync(directory: str) -> None:
    shutil.rmtree(directory)


async def file_exists(filePath: str) -> bool:
    return await aiofiles.os.path.exists(filePath)


def file_exists_sync(filePath: str) -> bool:
    return os.path.exists(filePath)


async def get_file_age_millis(filePath: str) -> int:
    # NOTE(krishan711): is there an async verison of this?
    return int((time.time() - os.path.getmtime(filePath)) * 1000)


def get_file_age_millis_sync(filePath: str) -> int:
    return int((time.time() - os.path.getmtime(filePath)) * 1000)


async def read_file(filePath: str) -> str:
    async with aiofiles.open(filePath, 'r') as file:
        return await file.read()


async def read_file_bytes(filePath: str) -> bytes:
    async with aiofiles.open(filePath, 'rb') as file:
        return await file.read()


def read_file_sync(filePath: str) -> str:
    with open(filePath, 'r') as file:
        return file.read()


def read_file_bytes_sync(filePath: str) -> bytes:
    with open(filePath, 'rb') as file:
        return file.read()


async def write_file(filePath: str, content: str, shouldCreateParentDirectories: bool = True, shouldRaiseIfFileExists: bool = False) -> None:
    if shouldCreateParentDirectories:
        await create_directory(directory=os.path.dirname(filePath), shouldAllowExisting=True)
    async with aiofiles.open(filePath, 'x' if shouldRaiseIfFileExists else 'w') as file:
        await file.write(content)


async def write_file_bytes(filePath: str, content: bytes, shouldCreateParentDirectories: bool = True, shouldRaiseIfFileExists: bool = False) -> None:
    if shouldCreateParentDirectories:
        await create_directory(directory=os.path.dirname(filePath), shouldAllowExisting=True)
    async with aiofiles.open(filePath, 'xb' if shouldRaiseIfFileExists else 'wb') as file:
        await file.write(content)


def write_file_sync(filePath: str, content: str, shouldCreateParentDirectories: bool = True, shouldRaiseIfFileExists: bool = False) -> None:
    if shouldCreateParentDirectories:
        create_directory_sync(directory=os.path.dirname(filePath), shouldAllowExisting=True)
    with open(filePath, 'x' if shouldRaiseIfFileExists else 'w') as file:
        file.write(content)


def write_file_bytes_sync(filePath: str, content: bytes, shouldCreateParentDirectories: bool = True, shouldRaiseIfFileExists: bool = False) -> None:
    if shouldCreateParentDirectories:
        create_directory_sync(directory=os.path.dirname(filePath), shouldAllowExisting=True)
    with open(filePath, 'xb' if shouldRaiseIfFileExists else 'wb') as file:
        file.write(content)


async def create_directory(directory: str, shouldAllowExisting: bool = True) -> None:
    pathlib.Path(directory).mkdir(parents=True, exist_ok=shouldAllowExisting)


def create_directory_sync(directory: str, shouldAllowExisting: bool = True) -> None:
    pathlib.Path(directory).mkdir(parents=True, exist_ok=shouldAllowExisting)


def get_file_extension(filename: str) -> str:
    _, fileExtension = os.path.splitext(filename)
    return fileExtension



from abc import ABC
from abc import abstractmethod


class Cache(ABC):
    def __init__(self, isPrivate: bool = False) -> None:
        super().__init__()
        self.isPrivate = isPrivate

    @abstractmethod
    async def set(self, key: str, value: str, expirySeconds: float) -> bool: ...

    @abstractmethod
    async def get(self, key: str) -> str | None: ...

    @abstractmethod
    async def delete(self, key: str) -> bool:
        """
        Returned bool indicates whether the record was removed or not
        """
        ...

    @abstractmethod
    def can_store_complex_objects(self) -> bool: ...



from __future__ import annotations

import base64
from binascii import Error as BinasciiError

from core.exceptions import UnauthorizedException


class BasicAuthentication:
    def __init__(self, username: str, password: str) -> None:
        self.username = username
        self.password = password

    @classmethod
    def from_string(cls, basicAuthenticationString: str) -> BasicAuthentication:
        try:
            decodedString = base64.b64decode(s=basicAuthenticationString.encode('utf-8'))
        except BinasciiError as exception:
            raise UnauthorizedException(message='Failed to decode BasicAuthentication value.') from exception
        parts = decodedString.split(b':')
        return cls(username=parts[0].decode('latin1'), password=parts[1].decode('latin1'))

    def to_string(self) -> str:
        basicAuthenticationString = b':'.join((self.username.encode('latin1'), self.password.encode('latin1')))
        return base64.b64encode(s=basicAuthenticationString).decode('utf-8')



from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.responses import Response
from starlette.routing import BaseRoute
from starlette.routing import Route


def create_default_routes(name: str, version: str, environment: str) -> list[BaseRoute]:
    async def root(request: Request) -> Response:  # noqa: ARG001
        return JSONResponse(content={'server': name, 'version': version, 'environment': environment})

    return [
        Route('/', methods=['GET'], endpoint=root),
    ]



from core.notifications.notification_client import NotificationClient
from core.requester import KibaResponse
from core.requester import Requester


class SlackClient(NotificationClient):
    def __init__(self, webhookUrl: str, requester: Requester, defaultChannel: str, defaultSender: str | None = 'kiba-server', defaultIconEmoji: str | None = ':robot_face:') -> None:
        self.webhookUrl = webhookUrl
        self.requester = requester
        self.defaultChannel = defaultChannel
        self.defaultSender = defaultSender
        self.defaultIconEmoji = defaultIconEmoji

    async def post(self, messageText: str) -> KibaResponse:
        response = await self.requester.post_json(
            url=self.webhookUrl,
            dataDict={
                'text': messageText,
                'username': self.defaultSender,
                'channel': self.defaultChannel,
                'icon_emoji': self.defaultIconEmoji,
            },
        )
        return response



from __future__ import annotations

from collections.abc import Mapping

from core.util.typing_util import Json


class KibaException(Exception):  # noqa: N818
    def __init__(self, message: str | None, statusCode: int | None = None, exceptionType: str | None = None) -> None:
        super().__init__(message)
        self.message = message
        self.statusCode = statusCode or 500
        self.exceptionType = exceptionType if exceptionType else self.__class__.__name__

    @staticmethod
    def from_exception(exception: Exception, statusCode: int = 500) -> KibaException:
        if isinstance(exception, KibaException):
            return exception
        return KibaException(message=str(exception), statusCode=statusCode, exceptionType=exception.__class__.__name__)

    def to_dict(self) -> dict[str, Json]:
        return {
            'exceptionType': self.exceptionType,
            'message': self.message,
            'fields': {},
            'statusCode': self.statusCode,
        }

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(statusCode={self.statusCode!r}, exceptionType={self.exceptionType!r}, message={self.message!r})'

    def __str__(self) -> str:
        return self.__repr__()

    def __eq__(self, other: object) -> bool:
        return isinstance(other, KibaException) and self.statusCode == other.statusCode and self.exceptionType == other.exceptionType

    def __hash__(self) -> int:
        return hash((self.statusCode, self.exceptionType))


class ClientException(KibaException):
    pass


class BadRequestException(ClientException):
    def __init__(self, message: str | None = None) -> None:
        message = message if message else 'Bad Request'
        super().__init__(message=message, statusCode=400)


class UnauthorizedException(ClientException):
    def __init__(self, message: str | None = None) -> None:
        message = message if message else 'Unauthorized'
        super().__init__(message=message, statusCode=401)


class ForbiddenException(ClientException):
    def __init__(self, message: str | None = None) -> None:
        message = message if message else 'Forbidden'
        super().__init__(message=message, statusCode=403)


class NotFoundException(ClientException):
    def __init__(self, message: str | None = None) -> None:
        message = message if message else 'Not Found'
        super().__init__(message=message, statusCode=404)


class ServerException(KibaException):
    pass


class InternalServerErrorException(ServerException):
    def __init__(self, message: str | None = None) -> None:
        message = message if message else 'Internal Server Error'
        super().__init__(message=message, statusCode=500)


class DuplicateValueException(BadRequestException):
    pass


class RedirectException(KibaException):
    def __init__(self, location: str, statusCode: int, message: str | None = None, shouldAddCacheHeader: bool | None = True) -> None:
        message = message if message else 'RedirectException'
        super().__init__(message=message, statusCode=statusCode)
        self.location = location
        self.shouldAddCacheHeader = shouldAddCacheHeader


class MovedPermanentlyRedirectException(RedirectException):
    def __init__(self, location: str, message: str | None = None, shouldAddCacheHeader: bool | None = True) -> None:
        message = message if message else 'Moved Permanently'
        super().__init__(location=location, message=message, statusCode=301, shouldAddCacheHeader=shouldAddCacheHeader)


class FoundRedirectException(RedirectException):
    def __init__(self, location: str, message: str | None = None) -> None:
        message = message if message else 'Found Redirect'
        super().__init__(location=location, message=message, statusCode=302, shouldAddCacheHeader=False)


class PermanentRedirectException(RedirectException):
    def __init__(self, location: str, message: str | None = None, shouldAddCacheHeader: bool | None = True) -> None:
        message = message if message else 'Permanent Redirect'
        super().__init__(location=location, message=message, statusCode=308, shouldAddCacheHeader=shouldAddCacheHeader)


CLIENT_EXCEPTIONS = [
    BadRequestException,
    ClientException,
    ForbiddenException,
    NotFoundException,
    UnauthorizedException,
    # ConflictException,
    # ExpectationFailedException,
    # FailedDependencyException,
    # GoneException,
    # ImATeapotException,
    # LengthRequiredException,
    # LockedException,
    # MethodNotAllowedException,
    # MisdirectedRequestException,
    # NotAcceptableException,
    # PayloadTooLargeException,
    # PaymentRequiredException,
    # PreconditionFailedException,
    # PreconditionRequiredException,
    # ProxyAuthenticationRequiredException,
    # RangeNotSatisfiableException,
    # RequestHeaderFieldsTooLargeException,
    # RequestTimeoutException,
    # TooManyRequestsException,
    # UnavailableForLegalReasonsException,
    # UnprocessableEntityException,
    # UnsupportedMediaTypeException,
    # UpgradeRequiredException,
    # UriTooLongException,
    # NoResponseException,
    # RequestHeaderTooLargeException,
    # SSLCertificateErrorException,
    # SSLCertificateRequiredException,
    # HttpRequestSentToHttpsException,
    # ClientClosedRequestException,
]

SERVER_EXCEPTIONS = [
    InternalServerErrorException,
    ServerException,
    # BadGatewayException,
    # BandwidthLimitExceededException,
    # ConnectionTimeoutException,
    # GatewayTimeoutException,
    # HttpVersionNotSupportedException,
    # InsufficientStorageException,
    # InvalidSSLCertificateException,
    # LoopDetectedException,
    # NetworkAuthenticationRequiredException,
    # NetworkReadTimeoutException,
    # NotExtendedException,
    # NotImplementedException,
    # OriginDNSErrorException,
    # RailgunErrorException,
    # ServiceUnavailableException,
    # SSLHandshakeFailedException,
    # TimeoutOccurredException,
    # UnknownErrorException,
    # UnreachableOriginException,
    # VariantAlsoNegotiatesException,
    # WebServerIsDownException,
]

CLIENT_EXCEPTIONS_MAP = {
    400: BadRequestException,
    401: UnauthorizedException,
    403: ForbiddenException,
    404: NotFoundException,
    # 402: PaymentRequiredException,
    # 405: MethodNotAllowedException,
    # 406: NotAcceptableException,
    # 407: ProxyAuthenticationRequiredException,
    # 408: RequestTimeoutException,
    # 409: ConflictException,
    # 410: GoneException,
    # 411: LengthRequiredException,
    # 412: PreconditionFailedException,
    # 413: PayloadTooLargeException,
    # 414: UriTooLongException,
    # 415: UnsupportedMediaTypeException,
    # 416: RangeNotSatisfiableException,
    # 417: ExpectationFailedException,
    # 418: ImATeapotException,
    # 421: MisdirectedRequestException,
    # 422: UnprocessableEntityException,
    # 423: LockedException,
    # 424: FailedDependencyException,
    # 426: UpgradeRequiredException,
    # 428: PreconditionRequiredException,
    # 429: TooManyRequestsException,
    # 431: RequestHeaderFieldsTooLargeException,
    # 444: NoResponseException,
    # 451: UnavailableForLegalReasonsException,
    # 494: RequestHeaderTooLargeException,
    # 495: SSLCertificateErrorException,
    # 496: SSLCertificateRequiredException,
    # 497: HttpRequestSentToHttpsException,
    # 499: ClientClosedRequestException,
}

SERVER_EXCEPTIONS_MAP = {
    500: InternalServerErrorException,
    # 501: NotImplementedException,
    # 502: BadGatewayException,
    # 503: ServiceUnavailableException,
    # 504: GatewayTimeoutException,
    # 505: HttpVersionNotSupportedException,
    # 506: VariantAlsoNegotiatesException,
    # 507: InsufficientStorageException,
    # 508: LoopDetectedException,
    # 509: BandwidthLimitExceededException,
    # 510: NotExtendedException,
    # 511: NetworkAuthenticationRequiredException,
    # 520: UnknownErrorException,
    # 521: WebServerIsDownException,
    # 522: ConnectionTimeoutException,
    # 523: UnreachableOriginException,
    # 524: TimeoutOccurredException,
    # 525: SSLHandshakeFailedException,
    # 526: InvalidSSLCertificateException,
    # 527: RailgunErrorException,
    # 530: OriginDNSErrorException,
    # 598: NetworkReadTimeoutException,
}

HTTP_EXCEPTIONS = CLIENT_EXCEPTIONS + SERVER_EXCEPTIONS
HTTP_EXCEPTIONS_MAP: Mapping[int, type[ServerException | ClientException]] = {**SERVER_EXCEPTIONS_MAP, **CLIENT_EXCEPTIONS_MAP}

ALL_EXCEPTION_CLASSES = [
    # DataException,
    # IntegrityException,
    # InvalidApiParameterException,
    # InvalidEncodingException,
    # InvalidParameterException,
    # InvalidUrlException,
    # MessageToBeRequeuedException,
    # OperationalException,
    # ResourceAlreadyExistsException,
    # UrlJoinException,
    # TimeoutException,
    # LockTimeoutException,
    # CleaningUrlException,
    # MissingExecutorException,
    # DuplicateJobTypeException,
    # JobNotFoundException,
    # TypeNotFoundException,
    # SignalTimeoutAlreadySetException,
    # SqsSendingFaultException,
    *HTTP_EXCEPTIONS
]



import contextlib
from collections.abc import AsyncIterator
from typing import TYPE_CHECKING
from typing import Any

from sqlalchemy import Table
from sqlalchemy.engine import Result
from sqlalchemy.sql.selectable import TypedReturnsRows

from core.exceptions import InternalServerErrorException
from core.store.database import Database
from core.store.database import DatabaseConnection
from core.store.database import ResultType

if TYPE_CHECKING:
    from sqlalchemy.sql._typing import _ColumnExpressionArgument
    from sqlalchemy.sql._typing import _DMLColumnArgument

    WhereClause = _ColumnExpressionArgument[bool]
else:
    _DMLColumnArgument = Any
    WhereClause = Any


CreateRecordValuesDict = dict[_DMLColumnArgument, Any]  # type: ignore[explicit-any]
UpdateRecordValuesDict = dict[_DMLColumnArgument, Any]  # type: ignore[explicit-any]


class SavingException(InternalServerErrorException):
    pass


class Saver:
    def __init__(self, database: Database) -> None:
        self.database = database

    @contextlib.asynccontextmanager
    async def create_transaction(self) -> AsyncIterator[DatabaseConnection]:
        async with self.database.create_transaction() as connection:
            yield connection

    async def _execute(self, query: TypedReturnsRows[ResultType], connection: DatabaseConnection | None = None) -> Result[ResultType]:
        try:
            if connection:
                return await self.database.execute(query=query, connection=connection)
            async with self.create_transaction() as newConnection:
                return await self.database.execute(query=query, connection=newConnection)
        # NOTE(krishan711): this could probs be more specific e.g. sqlalchemy.dialects.postgresql.asyncpg.IntegrityError
        except Exception as exception:
            raise SavingException(message=f'Error running save operation: {exception!s}') from exception

    async def _insert_record(self, table: Table, values: CreateRecordValuesDict, connection: DatabaseConnection | None = None) -> int:
        query = table.insert().values(values).returning(table.c.id)
        result = await self._execute(query=query, connection=connection)
        rowId = int(result.scalar_one())
        return rowId

    async def _update_records(self, table: Table, where: WhereClause, values: UpdateRecordValuesDict, connection: DatabaseConnection | None = None) -> list[int]:
        query = table.update().where(where).values(values).returning(table.c.id)
        result = await self._execute(query=query, connection=connection)
        rowIds = [int(rowId) for rowId in result.scalars()]
        return rowIds

    async def _delete_records(self, table: Table, where: WhereClause, connection: DatabaseConnection | None = None) -> list[int]:
        query = table.delete().where(where).returning(table.c.id)
        result = await self._execute(query=query, connection=connection)
        rowIds = [int(rowId) for rowId in result.scalars()]
        return rowIds



from core.util import url_util


class TestUrlUtil:

    def test_encode_url_with_special_characters(self):
        url = "https://example.com/path with spaces/file?param=value&special=!@#$%^"
        result = url_util.encode_url(url=url)
        assert result == "https%3A//example.com/path%20with%20spaces/file%3Fparam%3Dvalue%26special%3D%21%40%23%24%25%5E"

    def test_encode_url_with_unicode_characters(self):
        url = "https://example.com/üñîçødé/文字"
        result = url_util.encode_url(url=url)
        assert result == "https%3A//example.com/%C3%BC%C3%B1%C3%AE%C3%A7%C3%B8d%C3%A9/%E6%96%87%E5%AD%97"

    def test_encode_url_with_safe_characters(self):
        url = "https://example.com/normal-path/file.txt"
        result = url_util.encode_url(url=url)
        assert result == "https%3A//example.com/normal-path/file.txt"

    def test_encode_url_part_with_special_characters(self):
        url_part = "file name with spaces!@#$%^&*()"
        result = url_util.encode_url_part(urlPart=url_part)
        assert result == "file%20name%20with%20spaces%21%40%23%24%25%5E%26%2A%28%29"

    def test_encode_url_part_with_unicode_characters(self):
        url_part = "üñîçødé文字"
        result = url_util.encode_url_part(urlPart=url_part)
        assert result == "%C3%BC%C3%B1%C3%AE%C3%A7%C3%B8d%C3%A9%E6%96%87%E5%AD%97"

    def test_encode_url_part_with_slashes(self):
        url_part = "path/to/file"
        result = url_util.encode_url_part(urlPart=url_part)
        assert result == "path%2Fto%2Ffile"

    def test_decode_url_with_encoded_special_characters(self):
        encoded_url = "https%3A//example.com/path%20with%20spaces/file%3Fparam%3Dvalue%26special%3D%21%40%23%24%25%5E"
        result = url_util.decode_url(url=encoded_url)
        assert result == "https://example.com/path with spaces/file?param=value&special=!@#$%^"

    def test_decode_url_with_encoded_unicode_characters(self):
        encoded_url = "https%3A//example.com/%C3%BC%C3%B1%C3%AE%C3%A7%C3%B8d%C3%A9/%E6%96%87%E5%AD%97"
        result = url_util.decode_url(url=encoded_url)
        assert result == "https://example.com/üñîçødé/文字"

    def test_decode_url_with_no_encoding(self):
        url = "https://example.com/normal-path/file.txt"
        result = url_util.decode_url(url=url)
        assert result == url

    def test_decode_url_part_with_encoded_special_characters(self):
        encoded_part = "file%20name%20with%20spaces%21%40%23%24%25%5E%26%2A%28%29"
        result = url_util.decode_url_part(urlPart=encoded_part)
        assert result == "file name with spaces!@#$%^&*()"

    def test_decode_url_part_with_encoded_unicode_characters(self):
        encoded_part = "%C3%BC%C3%B1%C3%AE%C3%A7%C3%B8d%C3%A9%E6%96%87%E5%AD%97"
        result = url_util.decode_url_part(urlPart=encoded_part)
        assert result == "üñîçødé文字"

    def test_decode_url_part_with_encoded_slashes(self):
        encoded_part = "path%2Fto%2Ffile"
        result = url_util.decode_url_part(urlPart=encoded_part)
        assert result == "path/to/file"

    def test_encode_decode_url_roundtrip(self):
        original_url = "https://example.com/path with spaces/üñîçødé/file?param=value&special=!@#$%^"
        encoded = url_util.encode_url(url=original_url)
        decoded = url_util.decode_url(url=encoded)
        assert decoded == original_url

    def test_encode_decode_url_part_roundtrip(self):
        original_part = "path/to/file with spaces and üñîçødé!@#$%^"
        encoded = url_util.encode_url_part(urlPart=original_part)
        decoded = url_util.decode_url_part(urlPart=encoded)
        assert decoded == original_part



from __future__ import annotations

from collections.abc import Sequence
from contextlib import AsyncExitStack
from typing import TYPE_CHECKING
from typing import Any

from aiobotocore.session import get_session as get_botocore_session

from core.exceptions import InternalServerErrorException
from core.queues.message_queue import MessageQueue
from core.queues.model import Message
from core.util import list_util

if TYPE_CHECKING:
    from types_aiobotocore_sqs import SQSClient
    from types_aiobotocore_sqs.type_defs import MessageTypeDef as RawSqsMessageTypeDef
    from types_aiobotocore_sqs.type_defs import SendMessageBatchRequestEntryTypeDef
else:
    SQSClient = Any
    SendMessageBatchRequestEntryTypeDef = Any
    RawSqsMessageTypeDef = Any


class SqsMessage(Message):  # type: ignore[explicit-any]
    receiptHandle: str

    @classmethod
    def from_sqs_message(cls, sqsMessage: RawSqsMessageTypeDef) -> SqsMessage:
        message = Message.parse_raw(sqsMessage['Body'])
        return cls(
            command=message.command,
            content=message.content,
            requestId=message.requestId,
            postCount=message.postCount,
            postDate=message.postDate,
            receiptHandle=sqsMessage['ReceiptHandle'],
        )


class SqsMessageQueue(MessageQueue[SqsMessage]):
    def __init__(self, region: str, accessKeyId: str, accessKeySecret: str, queueUrl: str) -> None:
        self.region = region
        self._accessKeyId = accessKeyId
        self._accessKeySecret = accessKeySecret
        self.queueUrl = queueUrl
        self._exitStack = AsyncExitStack()
        self._sqsClient: SQSClient | None = None

    async def connect(self) -> None:
        session = get_botocore_session()
        self._sqsClient = await self._exitStack.enter_async_context(session.create_client('sqs', region_name=self.region, aws_access_key_id=self._accessKeyId, aws_secret_access_key=self._accessKeySecret))

    async def disconnect(self) -> None:
        await self._exitStack.aclose()
        self._sqsClient = None

    async def send_message(self, message: Message, delaySeconds: int = 0) -> None:
        if not self._sqsClient:
            raise InternalServerErrorException('You need to call .connect() before trying to send messages')
        message.prepare_for_send()
        await self._sqsClient.send_message(QueueUrl=self.queueUrl, DelaySeconds=delaySeconds, MessageAttributes={}, MessageBody=message.json())

    async def send_messages(self, messages: Sequence[Message], delaySeconds: int = 0) -> None:
        if not self._sqsClient:
            raise InternalServerErrorException('You need to call .connect() before trying to send messages')
        failures = []
        for messageChunk in list_util.generate_chunks(lst=messages, chunkSize=10):
            requests: list[SendMessageBatchRequestEntryTypeDef] = []
            for index, message in enumerate(messageChunk):
                message.prepare_for_send()
                requests.append({'Id': str(index), 'DelaySeconds': int(delaySeconds), 'MessageAttributes': {}, 'MessageBody': message.json()})
            response = await self._sqsClient.send_message_batch(QueueUrl=self.queueUrl, Entries=requests)
            failures += response.get('Failed', [])
        if len(failures) > 0:
            errorMessage = ''
            for failure in failures:
                failureType = 'Sender' if failure['SenderFault'] else 'Receiver'
                errorMessage += f'{failureType} fault: id {failure["Id"]}, code {failure["Code"]}, message {failure.get("Message")}\n'
            raise InternalServerErrorException(message=errorMessage)

    async def get_message(self, expectedProcessingSeconds: int = 300, longPollSeconds: int = 0) -> SqsMessage | None:
        messages = await self.get_messages(limit=1, expectedProcessingSeconds=expectedProcessingSeconds, longPollSeconds=longPollSeconds)
        return messages[0] if messages else None

    async def get_messages(self, limit: int = 1, expectedProcessingSeconds: int = 300, longPollSeconds: int = 0) -> list[SqsMessage]:
        if not self._sqsClient:
            raise InternalServerErrorException('You need to call .connect() before trying to get messages')
        sqsResponse = await self._sqsClient.receive_message(QueueUrl=self.queueUrl, VisibilityTimeout=expectedProcessingSeconds, MaxNumberOfMessages=limit, WaitTimeSeconds=longPollSeconds)
        sqsMessages = [SqsMessage.from_sqs_message(sqsMessage=sqsMessage) for sqsMessage in sqsResponse.get('Messages', [])]
        return sqsMessages

    async def delete_message(self, message: SqsMessage) -> None:
        if not self._sqsClient:
            raise InternalServerErrorException('You need to call .connect() before trying to delete messages')
        await self._sqsClient.delete_message(QueueUrl=self.queueUrl, ReceiptHandle=message.receiptHandle)



from collections.abc import Iterator
from collections.abc import Sequence
from typing import TypeVar

ListItemType = TypeVar('ListItemType')


def generate_chunks(lst: Sequence[ListItemType], chunkSize: int) -> Iterator[Sequence[ListItemType]]:
    for index in range(0, len(lst), chunkSize):
        yield lst[index : index + chunkSize]



from core.util import list_util


class TestListUtil:

    def test_generate_chunks_with_even_division(self):
        inputList = [1, 2, 3, 4, 5, 6]
        chunkSize = 2
        expected = [[1, 2], [3, 4], [5, 6]]
        result = list(list_util.generate_chunks(inputList, chunkSize))
        assert result == expected

    def test_generate_chunks_with_uneven_division(self):
        inputList = [1, 2, 3, 4, 5]
        chunkSize = 2
        expected = [[1, 2], [3, 4], [5]]
        result = list(list_util.generate_chunks(inputList, chunkSize))
        assert result == expected

    def test_generate_chunks_with_empty_list(self):
        inputList: list[int] = []
        chunkSize = 2
        expected = []
        result = list(list_util.generate_chunks(inputList, chunkSize))
        assert result == expected

    def test_generate_chunks_with_chunk_size_larger_than_list(self):
        inputList = [1, 2, 3]
        chunkSize = 5
        expected = [[1, 2, 3]]
        result = list(list_util.generate_chunks(inputList, chunkSize))
        assert result == expected

    def test_generate_chunks_with_different_types(self):
        inputList = ["a", "b", "c", "d"]
        chunkSize = 3
        expected = [["a", "b", "c"], ["d"]]
        result = list(list_util.generate_chunks(inputList, chunkSize))
        assert result == expected



import binascii
import hashlib


def sha512(value: str, salt: str = '', iterationCount: int = 1000) -> str:
    return binascii.hexlify(hashlib.pbkdf2_hmac(hash_name='sha512', password=value.encode(), salt=salt.encode(), iterations=iterationCount)).decode()



from core.util import dict_util


class TestDictUtil:

    def test_merge_dicts_with_non_overlapping_keys(self):
        dict1 = {'a': 1, 'b': 2}
        dict2 = {'c': 3, 'd': 4}
        expected = {'a': 1, 'b': 2, 'c': 3, 'd': 4}
        result = dict_util.merge_dicts(dict1, dict2)
        assert result == expected

    def test_merge_dicts_with_overlapping_keys(self):
        dict1 = {'a': 1, 'b': 2}
        dict2 = {'b': 3, 'c': 4}
        expected = {'a': 1, 'b': 3, 'c': 4}  # dict2 values should override dict1 values
        result = dict_util.merge_dicts(dict1, dict2)
        assert result == expected

    def test_merge_dicts_with_empty_dicts(self):
        dict1: dict[str, int] = {}
        dict2: dict[str, int] = {}
        expected: dict[str, int] = {}
        result = dict_util.merge_dicts(dict1, dict2)
        assert result == expected

    def test_merge_dicts_with_one_empty_dict(self):
        dict1 = {'a': 1, 'b': 2}
        dict2: dict[str, int] = {}
        expected = {'a': 1, 'b': 2}
        result = dict_util.merge_dicts(dict1, dict2)
        assert result == expected

    def test_merge_dicts_with_different_value_types(self):
        dict1 = {'a': 1, 'b': 'string'}
        dict2 = {'c': True, 'd': 2.5}
        expected = {'a': 1, 'b': 'string', 'c': True, 'd': 2.5}
        result = dict_util.merge_dicts(dict1, dict2)
        assert result == expected



from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.base import RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.responses import Response

from core import logging
from core.exceptions import KibaException
from core.exceptions import RedirectException


class ExceptionHandlingMiddleware(BaseHTTPMiddleware):
    @staticmethod
    def _convert_exception(exception: KibaException) -> Response:
        response = JSONResponse(status_code=exception.statusCode, content=exception.to_dict())
        return response

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        try:
            response = await call_next(request)
        except RedirectException as exception:
            response = Response(status_code=exception.statusCode)
            response.headers['Location'] = exception.location
            # TODO(krishan711): clean this up after major version bump (since shouldAddCacheHeader field will always exist)
            if not hasattr(exception, 'shouldAddCacheHeader') or exception.shouldAddCacheHeader:
                response.headers['Cache-Control'] = f'max-age={60 * 60 * 24 * 365}'
        except KibaException as exception:
            logging.exception(exception)
            response = self._convert_exception(exception=exception)
        except Exception as exception:  # noqa: BLE001
            logging.exception(exception)
            response = self._convert_exception(exception=KibaException.from_exception(exception=exception))
        return response



from __future__ import annotations

import asyncio
import os
from threading import Lock

from core.caching.cache import Cache
from core.util import date_util
from core.util import file_util


class FileCache(Cache):
    def __init__(self, cacheDirectory: str, isPrivate: bool = False) -> None:
        super().__init__(isPrivate=isPrivate)
        self._cacheDirectory = cacheDirectory
        self._lock = Lock()

    # async def load_cache_file(
    #     fileName: str,
    #     expirySeconds: int = 3600,
    # ) -> str | None:
    #     cacheFileDirectory = os.path.join('./data/cache', fileName)
    #     fileExists = await file_util.file_exists(filePath=cacheFileDirectory)
    #     fileAgeMillis = await file_util.get_file_age_millis(filePath=cacheFileDirectory) if fileExists else 0
    #     if fileExists and (fileAgeMillis / 1000) < expirySeconds:
    #         return await file_util.read_file(filePath=cacheFileDirectory)
    #     return None

    # async def load_json_cache_file(  # type: ignore[explicit-any]
    #     fileName: str,
    #     expirySeconds: int = 3600,
    # ) -> dict[str, Any] | list[Any] | None:
    #     content = await load_cache_file(fileName=fileName, expirySeconds=expirySeconds)
    #     return json.loads(content) if content is not None else None

    async def set(self, key: str, value: str, expirySeconds: float) -> bool:
        cacheFileDirectory = os.path.join(self._cacheDirectory, key)
        contentFilePath = os.path.join(cacheFileDirectory, 'content.txt')
        expiryFilePath = os.path.join(cacheFileDirectory, 'expiryDate.txt')
        expiryDateString = date_util.datetime_to_string(dt=date_util.datetime_from_now(seconds=expirySeconds))
        with self._lock:
            await asyncio.gather(
                *[
                    file_util.write_file(filePath=contentFilePath, content=value),
                    file_util.write_file(filePath=expiryFilePath, content=expiryDateString),
                ]
            )
            return True

    async def _internal_get(self, key: str) -> str | None:
        cacheFileDirectory = os.path.join(self._cacheDirectory, key)
        contentFilePath = os.path.join(cacheFileDirectory, 'content.txt')
        expiryFilePath = os.path.join(cacheFileDirectory, 'expiryDate.txt')
        contentFileExists = await file_util.file_exists(filePath=contentFilePath)
        if not contentFileExists:
            return None
        expiryFileExists = await file_util.file_exists(filePath=expiryFilePath)
        if not expiryFileExists:
            await file_util.remove_file(filePath=contentFilePath)
            return None
        expiryDateString = await file_util.read_file(filePath=expiryFilePath)
        expiryDate = date_util.datetime_from_string(dateString=expiryDateString)
        if expiryDate < date_util.datetime_from_now():
            await file_util.remove_file(filePath=contentFilePath)
            return None
        return await file_util.read_file(filePath=contentFilePath)

    async def get(self, key: str) -> str | None:
        with self._lock:
            return await self._internal_get(key=key)

    async def delete(self, key: str) -> bool:
        # NOTE(krishan711): we only delete content file for speed
        cacheFileDirectory = os.path.join(self._cacheDirectory, key)
        contentFilePath = os.path.join(cacheFileDirectory, 'content.txt')
        with self._lock:
            fileExists = await file_util.file_exists(filePath=contentFilePath)
            if fileExists:
                await file_util.remove_file(filePath=contentFilePath)
            return fileExists

    def can_store_complex_objects(self) -> bool:
        return False



# noqa: A005
import dataclasses
import datetime
import json
import logging
import os
import re
import sys
import typing
from collections.abc import Callable
from collections.abc import Collection
from logging import Formatter
from logging import Logger
from logging import LogRecord
from typing import TYPE_CHECKING
from typing import Any
from typing import TextIO
from typing import Union

from core.exceptions import KibaException
from core.util.value_holder import RequestIdHolder

if TYPE_CHECKING:
    # NOTE(krishan711): this is only processed by mypy
    StreamHandler = logging.StreamHandler[TextIO]
else:
    from logging import StreamHandler


@dataclasses.dataclass
class LogFormat:
    loggerType: str
    loggerName: str
    loggerFormat: str
    jsonFieldFormatters: dict[str, Callable[[str], Union[str, int, float, None]]]


def json_parse_string_value(value: str) -> Union[str, None]:
    if value == '':
        return None
    return str(value)


def json_parse_int_value(value: str) -> Union[int, None]:
    if value == '':
        return None
    return int(value)


def json_parse_float_value(value: str) -> Union[float, None]:
    if value == '':
        return None
    return float(value)


_LOGGING_FORMAT_VERSION = 1
ROOT_FORMAT = LogFormat(loggerType='', loggerName=f'KIBA_{_LOGGING_FORMAT_VERSION}', loggerFormat='%(message)s', jsonFieldFormatters={})
STAT_FORMAT = LogFormat(
    loggerType='stat', loggerName=f'KIBA_STAT_{_LOGGING_FORMAT_VERSION}', loggerFormat='%(statName)s:%(statKey)s:%(statValue)s', jsonFieldFormatters={'statName': json_parse_string_value, 'statKey': json_parse_string_value, 'statValue': json_parse_float_value}
)
API_FORMAT = LogFormat(
    loggerType='api',
    loggerName=f'KIBA_API_{_LOGGING_FORMAT_VERSION}',
    loggerFormat='%(apiAction)s:%(apiPath)s:%(apiQuery)s:%(apiResponse)s:%(apiDuration)s',
    jsonFieldFormatters={'apiAction': json_parse_string_value, 'apiPath': json_parse_string_value, 'apiQuery': json_parse_string_value, 'apiResponse': json_parse_int_value, 'apiDuration': json_parse_float_value},
)
ALL_LOGGER_FORMATS = [ROOT_FORMAT, STAT_FORMAT, API_FORMAT]

ROOT_LOGGER = logging.getLogger(ROOT_FORMAT.loggerType)
STAT_LOGGER = logging.getLogger(STAT_FORMAT.loggerType)
API_LOGGER = logging.getLogger(API_FORMAT.loggerType)


class KibaLoggingFormatter(Formatter):
    _BASE_FORMAT_STRING = '%(asctime)s - {format} - {name} - {version} - {environment} - %(requestId)s - %(levelname)s - %(name)s - %(pathname)s:%(funcName)s:%(lineno)s - {logMessage}'

    def __init__(self, logFormat: LogFormat, name: str, version: str, environment: str, requestIdHolder: RequestIdHolder | None) -> None:
        formatString = self._BASE_FORMAT_STRING.format(format=logFormat.loggerName, name=name, version=version, environment=environment, logMessage=logFormat.loggerFormat)
        super().__init__(fmt=formatString)
        self.logFormat = logFormat
        self.name = name
        self.version = version
        self.environment = environment
        self.requestIdHolder = requestIdHolder

    def format(self, record: LogRecord) -> str:
        record.pathname = re.sub(os.getcwd() + '.', '', record.pathname).replace('/', '.')
        record.requestId = self.requestIdHolder.get_value() if self.requestIdHolder is not None else ''
        return super().format(record=record)

    def formatTime(self, record: LogRecord, datefmt: str | None = None) -> str:  # noqa: N802]
        logDate = datetime.datetime.fromtimestamp(record.created, tz=datetime.UTC)
        return logDate.strftime(datefmt or '%Y-%m-%dT%H:%M:%S.%f')


class KibaJsonLoggingFormatter(KibaLoggingFormatter):
    def format(self, record: LogRecord) -> str:
        message = self.formatException(record.exc_info) if record.exc_info else str(record.msg)
        recordDict: dict[str, Union[str, int, float, bool, None]] = {
            'date': self.formatTime(record, '%Y-%m-%dT%H:%M:%S.%f'),
            # NOTE(krishan711): for some reason these constantly report this file instead of original
            # 'path': record.pathname,
            # 'function': record.funcName,
            # 'line': record.lineno,
            'message': message or None,
            'level': record.levelname,
            'logger': record.name,
            'format': self.logFormat.loggerName,
            'name': self.name,
            'version': self.version,
            'environment': self.environment,
            'requestId': self.requestIdHolder.get_value() if self.requestIdHolder is not None else None,
        }
        for fieldName, formatter in self.logFormat.jsonFieldFormatters.items():
            recordDict[fieldName] = formatter(record.__dict__[fieldName])
        return json.dumps(recordDict)


def init_logger(logger: Logger, loggingLevel: int, handler: StreamHandler) -> None:
    logger.propagate = False
    logger.handlers = [handler]
    logger.setLevel(level=loggingLevel)


def init_external_loggers(loggerNames: Collection[str], loggingLevel: int = logging.WARNING) -> None:
    for loggerName in loggerNames:
        logging.getLogger(loggerName).setLevel(loggingLevel)


def init_logging(name: str, version: str, environment: str, showDebug: bool = False, requestIdHolder: RequestIdHolder | None = None) -> None:
    loggingLevel = logging.DEBUG if showDebug else logging.INFO
    for logFormat in ALL_LOGGER_FORMATS:
        logger = logging.getLogger(name=logFormat.loggerType)
        handler = StreamHandler(stream=sys.stdout)
        handler.setFormatter(fmt=KibaLoggingFormatter(logFormat=logFormat, name=name, version=version, environment=environment, requestIdHolder=requestIdHolder))
        init_logger(logger=logger, loggingLevel=loggingLevel, handler=handler)


def init_json_logging(name: str, version: str, environment: str, showDebug: bool = False, requestIdHolder: RequestIdHolder | None = None) -> None:
    loggingLevel = logging.DEBUG if showDebug else logging.INFO
    for logFormat in ALL_LOGGER_FORMATS:
        logger = logging.getLogger(name=logFormat.loggerType)
        handler = StreamHandler(stream=sys.stdout)
        handler.setFormatter(fmt=KibaJsonLoggingFormatter(logFormat=logFormat, name=name, version=version, environment=environment, requestIdHolder=requestIdHolder))
        init_logger(logger=logger, loggingLevel=loggingLevel, handler=handler)


def init_basic_logging(showDebug: bool = False) -> None:
    loggingLevel = logging.DEBUG if showDebug else logging.INFO
    for logFormat in ALL_LOGGER_FORMATS:
        logger = logging.getLogger(name=logFormat.loggerType)
        handler = StreamHandler(stream=sys.stdout)
        handler.setFormatter(fmt=Formatter(fmt=logFormat.loggerFormat))
        init_logger(logger=logger, loggingLevel=loggingLevel, handler=handler)


def _serialize_numeric_value(value: Union[None, float]) -> str:
    if value is None:
        return ''
    roundedNumber = round(value, 6)
    return f'{roundedNumber:f}'.rstrip('0').rstrip('.')


def _serialize_string_value(value: str) -> str:
    return value.replace(':', '__')


def stat(name: str, key: str, value: float = 1) -> None:
    if STAT_LOGGER.isEnabledFor(level=logging.INFO):
        nameValue = _serialize_string_value(value=str(name))
        keyValue = _serialize_string_value(value=str(key))
        statValue = _serialize_numeric_value(value=value)
        STAT_LOGGER.log(level=logging.INFO, msg='', extra=typing.cast(dict[str, str], {'statName': nameValue, 'statKey': keyValue, 'statValue': statValue}))


def api(action: str, path: str, query: str, response: int | None = None, duration: float | None = None) -> None:
    if API_LOGGER.isEnabledFor(level=logging.INFO):
        actionString = _serialize_string_value(value=action)
        pathString = _serialize_string_value(value=path)
        queryString = _serialize_string_value(value=query)
        responseString = _serialize_numeric_value(value=response)
        durationString = _serialize_numeric_value(value=duration)
        API_LOGGER.log(level=logging.INFO, msg='', extra=typing.cast(dict[str, str], {'apiAction': actionString, 'apiPath': pathString, 'apiQuery': queryString, 'apiResponse': responseString or '', 'apiDuration': durationString or ''}))


# Wrappers around common python logging functions which go straight to the root logger

CRITICAL = logging.CRITICAL
ERROR = logging.ERROR
ERROR = logging.ERROR
WARNING = logging.WARNING
INFO = logging.INFO
DEBUG = logging.DEBUG


def _log(level: int, msg: str, *args: Any, **kwargs: Any) -> None:  # type: ignore[explicit-any]
    if ROOT_LOGGER.isEnabledFor(level=level):
        ROOT_LOGGER._log(level=level, msg=msg, args=args, **kwargs)  # noqa: SLF001


def critical(msg: str, *args: Any, **kwargs: Any) -> None:  # type: ignore[explicit-any]
    _log(level=CRITICAL, msg=msg, *args, **kwargs)  # type: ignore[misc]  # noqa: B026


def error(msg: str, *args: Any, **kwargs: Any) -> None:  # type: ignore[explicit-any]
    _log(level=ERROR, msg=msg, *args, **kwargs)  # type: ignore[misc]  # noqa: B026


def exception(msg: Union[str, Exception], *args: Any, exc_info: bool = True, **kwargs: Any) -> None:  # type: ignore[explicit-any]
    if isinstance(msg, KibaException):
        resolvedMessage = msg.message or str(msg)
    elif isinstance(msg, Exception):
        resolvedMessage = str(msg)
    else:
        resolvedMessage = msg
    _log(level=ERROR, msg=resolvedMessage, exc_info=exc_info, *args, **kwargs)  # type: ignore[misc]  # noqa: B026


def warning(msg: str, *args: Any, **kwargs: Any) -> None:  # type: ignore[explicit-any]
    _log(level=WARNING, msg=msg, *args, **kwargs)  # type: ignore[misc]  # noqa: B026


def info(msg: str, *args: Any, **kwargs: Any) -> None:  # type: ignore[explicit-any]
    _log(level=INFO, msg=msg, *args, **kwargs)  # type: ignore[misc]  # noqa: B026


def debug(msg: str, *args: Any, **kwargs: Any) -> None:  # type: ignore[explicit-any]
    _log(level=DEBUG, msg=msg, *args, **kwargs)  # type: ignore[misc]  # noqa: B026


def basicConfig(**kwargs: Any) -> None:  # type: ignore[explicit-any]  # noqa: N802  # noqa: B026
    logging.basicConfig(**kwargs)


def getLogger(name: str | None = None) -> Logger:  # noqa: N802
    return logging.getLogger(name=name)



import json
import os

from pydantic import BaseModel

from core.requester import Requester


class ServiceClient:
    def __init__(self, requester: Requester, baseUrl: str) -> None:
        self.requester = requester
        self.baseUrl = baseUrl

    async def make_request(self, method: str, path: str, responseClass: type[BaseModel] | None = None, request: BaseModel | None = None) -> BaseModel | None:
        url = os.path.join(self.baseUrl, path)
        response = await self.requester.make_request(method=method, url=url, dataDict=request.model_dump() if request else None)
        return responseClass.model_validate(json.loads(response.content)) if responseClass else None



import typing

import orjson
from starlette.responses import JSONResponse


class KibaJSONResponse(JSONResponse):
    def render(self, content: typing.Any) -> bytes:  # type: ignore[explicit-any]
        return orjson.dumps(content)



from __future__ import annotations

import asyncio
from collections.abc import Sequence

from azure.storage.queue import QueueMessage as RawAqsMessage
from azure.storage.queue.aio import QueueClient

from core.exceptions import InternalServerErrorException
from core.queues.message_queue import MessageQueue
from core.queues.model import Message
from core.util import list_util


class AqsMessage(Message):  # type: ignore[explicit-any]
    aqsId: str
    popReceipt: str | None

    @classmethod
    def from_aqs_message(cls, aqsMessage: RawAqsMessage) -> AqsMessage:
        message = Message.parse_raw(aqsMessage.content)
        return cls(
            command=message.command,
            content=message.content,
            requestId=message.requestId,
            postCount=message.postCount,
            postDate=message.postDate,
            aqsId=aqsMessage.id,
            popReceipt=aqsMessage.pop_receipt,
        )


class AqsMessageQueue(MessageQueue[AqsMessage]):
    def __init__(self, storageAccountName: str, storageAccountKey: str, queueName: str) -> None:
        self._storageAccountName = storageAccountName
        self._storageAccountKey = storageAccountKey
        self.queueName = queueName
        self._aqsClient: QueueClient | None = None

    async def connect(self) -> None:
        self._aqsClient = QueueClient.from_connection_string(conn_str=f'DefaultEndpointsProtocol=https;AccountName={self._storageAccountName};AccountKey={self._storageAccountKey}', queue_name=self.queueName)
        if not self._aqsClient:
            raise InternalServerErrorException('Failed to connect to queue')
        await self._aqsClient.get_queue_properties()

    async def disconnect(self) -> None:
        if not self._aqsClient:
            return
        await self._aqsClient.close()  # type: ignore[no-untyped-call]
        self._aqsClient = None

    async def send_message(self, message: Message, delaySeconds: int = 0) -> None:
        if not self._aqsClient:
            raise InternalServerErrorException('You need to call .connect() before trying to send messages')
        message.prepare_for_send()
        await self._aqsClient.send_message(visibility_timeout=delaySeconds, content=message.json())

    async def send_messages(self, messages: Sequence[Message], delaySeconds: int = 0) -> None:
        if not self._aqsClient:
            raise InternalServerErrorException('You need to call .connect() before trying to send messages')
        for messageChunk in list_util.generate_chunks(lst=messages, chunkSize=10):
            await asyncio.gather(*[self.send_message(message=message, delaySeconds=delaySeconds) for message in messageChunk])

    async def get_message(self, expectedProcessingSeconds: int = 300, longPollSeconds: int = 0) -> AqsMessage | None:
        if not self._aqsClient:
            raise InternalServerErrorException('You need to call .connect() before trying to send messages')
        message = await self._aqsClient.receive_message(visibility_timeout=expectedProcessingSeconds, timeout=longPollSeconds)
        aqsMessage = AqsMessage.from_aqs_message(aqsMessage=message) if message else None
        return aqsMessage

    async def get_messages(self, limit: int = 1, expectedProcessingSeconds: int = 300, longPollSeconds: int = 0) -> list[AqsMessage]:
        if not self._aqsClient:
            raise InternalServerErrorException('You need to call .connect() before trying to get messages')
        messagesIterator = self._aqsClient.receive_messages(messages_per_page=min(limit, 10), max_messages=limit, visibility_timeout=expectedProcessingSeconds, timeout=longPollSeconds)
        aqsMessages = [AqsMessage.from_aqs_message(aqsMessage=message) async for message in messagesIterator]
        return aqsMessages

    async def delete_message(self, message: AqsMessage) -> None:
        if not self._aqsClient:
            raise InternalServerErrorException('You need to call .connect() before trying to delete messages')
        await self._aqsClient.delete_message(message=message.aqsId, pop_receipt=message.popReceipt)



from fastapi import APIRouter


def create_api(name: str, version: str, environment: str) -> APIRouter:
    router = APIRouter()

    @router.get('/')
    async def root() -> dict[str, str]:
        return {'server': name, 'version': version, 'environment': environment}

    return router



import datetime
import pytest

from core.caching.dict_cache import DictCache
from core.util import date_util


class TestDictCache:

    @pytest.fixture
    def cache(self) -> DictCache:
        return DictCache()

    async def test_set_and_get(self, cache: DictCache):
        key = "test_key"
        value = "test_value"
        success = await cache.set(key=key, value=value, expirySeconds=60)
        assert success is True
        result = await cache.get(key=key)
        assert result == value

    async def test_get_nonexistent_key(self, cache: DictCache):
        result = await cache.get(key="nonexistent")
        assert result is None

    async def test_delete_existing_key(self, cache: DictCache):
        key = "test_key"
        value = "test_value"
        await cache.set(key=key, value=value, expirySeconds=60)
        success = await cache.delete(key=key)
        assert success is True
        result = await cache.get(key=key)
        assert result is None

    async def test_delete_nonexistent_key(self, cache: DictCache):
        success = await cache.delete(key="nonexistent")
        assert success is False

    async def test_expiry(self, cache: DictCache, monkeypatch):
        key = "test_key"
        value = "test_value"
        await cache.set(key=key, value=value, expirySeconds=60)

        # Mock the current time to be after expiry
        future_time = date_util.datetime_from_now(seconds=61)
        monkeypatch.setattr(date_util, "datetime_from_now", lambda: future_time)

        result = await cache.get(key=key)
        assert result is None

    async def test_overwrite_existing_key(self, cache: DictCache):
        key = "test_key"
        value1 = "test_value_1"
        value2 = "test_value_2"
        await cache.set(key=key, value=value1, expirySeconds=60)
        await cache.set(key=key, value=value2, expirySeconds=60)
        result = await cache.get(key=key)
        assert result == value2

    def test_can_store_complex_objects(self, cache: DictCache):
        assert cache.can_store_complex_objects() is True

    def test_is_private_flag(self):
        private_cache = DictCache(isPrivate=True)
        assert private_cache.isPrivate is True
        public_cache = DictCache(isPrivate=False)
        assert public_cache.isPrivate is False
